import abc
from typing import *


class NEVectorization(metaclass=abc.ABCMeta):
    """
    The process of converting text into vector is called vectorization.
    The set of corpus documents vectorized corpus makes up the Vector Space Model, which can have a sparse or
    dense representation.

    A NEVectorization object allows you to create vectors grouping named_entities for arbitrary fields.

    Grouping named_entities by:

       * _doc_id_: you 're creating document vectors;
       * _sent_id_: you 're creating sentence vectors;
       * _author_: you're creating author vectors (each named_entities must have an 'author' field);
       ...

    You can group by multiple fields.

    It offers several functions to set the vector components values, such as:

       * **One-hot encoding** (binary values 0/1);
       * **Tf** (Term Frequency);
       * **TfIdf** (Term Frequency Inverse Document Frequency);
       * **MI** (Mutual Information)

    You can specify the vectorization representation format: Term x Document matrix, Postings list.
    """

    def __init__(self, group_by_field: Union[str, List[str]], ne_field: str = 'named_entities') -> None:
        """
        **\_\_init\_\_**: Creates a new Vectorization object.

        Args:
            group_by_field (Union[str, List[str]]): Fields linked to named_entities in the corpus (e.g. doc_id, sent_id, metadata1, ...).  <br/> The named_entities will be grouped by these field.
            ne_field (str, optional, default=named_entities): A field containing the named_entities list.
        """

        self.ne_field = ne_field + '_'

        if isinstance(group_by_field, str):
            self.group_by_field = [group_by_field + '_']
        else:
            self.group_by_field = [g + '_' for g in group_by_field]

        self.group_by_field_key = " + ".join(self.group_by_field)

    @abc.abstractmethod
    def tf(self, index: Iterable[dict],
           min_len: int = 0,
           min_count: int = 0,
           format_: str = "docterm-matrix",
           boolean=False) -> Union[Dict[str, dict], Iterable[dict]]:
        """
        Calculates Term Frequency.

        Args:
            index (Iterable[dict]): Records of an index produced during the tokenization stage (e.g. sentences, paragraphs, documents).

                    Example:
                    >>> sentence = next(sentences)
                    >>> print(sentence)
                    {
                    "sent_id_": 0,
                    "text_": "When We , though all unworthy , were called to succeed on the Apostolic Throne the meek Pius X [...]",
                    "lemma_": "when -PRON- , though all unworthy , be call to succeed on the Apostolic Throne the meek Pius X [ ... ]",
                    [...]
                    "named_entities_": [{"text": "Apostolic Throne", "start":12, "end":13}, {"text": "Pius X", "start": 16, "end":17}],
                    ...
                    }

            min_len (int, optional, default=0): The minimum length a named_entity must have to be considered. A named_entity must be at least min_len characters long.
            min_count (int, optional, default=0): The minimum frequency a named_entity must have in the corpus to be considered.
            format_ (str, optional, default=docterm-matrix): Format of the vectorization records.

                    Valid input values:
                    ["postings-list", "docterm-matrix"]

            boolean (bool, optional, default=False): If True return Boolean Vectors else Term Frequency. <br/> If True, the result is the same as onehot method.

        Returns:
            Union[Dict[str, dict], Iterable[dict]]: If format_ == 'postings-list', returns a postings list; else a docterm matrix.

                    Example:
                    [format_ == "postings-list"]

                    {
                        "named_entity1":
                              {
                                "doc1": tf1,
                                ...
                              },
                        ...
                    }

                     Example:
                    [format_ == "docterm-matrix"]

                    [{
                        "doc": "doc1",
                        "bow":
                              {
                               named_entity1: tf1,
                               ...
                               }
                    },
                    ...
                    ]
        """

    @abc.abstractmethod
    def tfidf(self,
              index: Iterable[dict],
              normalize=True,
              min_len=0,
              min_count: int = 0,
              format_: str = 'docterm-matrix') -> Union[Dict[str, dict], Iterable[dict]]:
        """
        Calculates Term Frequency Inverse Document Frequency.


        Args:
            index (Iterable[dict]): Records of an index produced during the tokenization stage (e.g. sentences, paragraphs, documents).

                    Example:
                    >>> sentence = next(sentences)
                    >>> print(sentence)
                    {
                    "sent_id_": 0,
                    "text_": "When We , though all unworthy , were called to succeed on the Apostolic Throne the meek Pius X [...]",
                    "lemma_": "when -PRON- , though all unworthy , be call to succeed on the Apostolic Throne the meek Pius X [ ... ]",
                    [...]
                    "named_entities_": [{"text": "Apostolic Throne", "start":12, "end":13}, {"text": "Pius X", "start": 16, "end":17}],
                    ...
                    }

            normalize (bool, optional, default=True): If True, normalizes TF for max(TF).
            min_len (int, optional, default=0): The minimum length a named_entity must have to be considered. A named_entity must be at least min_len characters long.
            min_count (int, optional, default=0): The minimum frequency a named_entity must have in the corpus to be considered.
            format_ (str, optional, default=docterm-matrix): Format of the vectorization records.

                    Valid input values:
                    ["postings-list", "docterm-matrix"]

        Returns:
            Union[Dict[str, dict], Iterable[dict]]: If format_ == 'postings-list', returns a postings list; else a docterm matrix.

                    Example:
                    [format_ == "postings-list"]

                    {
                        "named_entity1":
                              {
                                "doc1": tfidf1,
                                ...
                              },
                        ...
                    }

                     Example:
                    [format_ == "docterm-matrix"]

                    [{
                        "doc": "doc1",
                        "bow":
                              {
                               named_entity1: tfidf1,
                               ...
                               }
                    },
                    ...
                    ]
        """

    @abc.abstractmethod
    def mi(self,
           index: Iterable[dict],
           min_len: int = 0,
           min_count: int = 0,
           format_: str = 'docterm-matrix') -> Union[Dict[str, dict], Iterator[dict]]:
        """
        Calculates Mutual Information.

        Args:
            index (Iterable[dict]): Records of an index produced during the tokenization stage (e.g. sentences, paragraphs, documents).

                    Example:
                    >>> sentence = next(sentences)
                    >>> print(sentence)
                    {
                    "sent_id_": 0,
                    "text_": "When We , though all unworthy , were called to succeed on the Apostolic Throne the meek Pius X [...]",
                    "lemma_": "when -PRON- , though all unworthy , be call to succeed on the Apostolic Throne the meek Pius X [ ... ]",
                    [...]
                    "named_entities_": [{"text": "Apostolic Throne", "start":12, "end":13}, {"text": "Pius X", "start": 16, "end":17}],
                    ...
                    }

            min_len (int, optional, default=0): The minimum length a named_entity must have to be considered. A named_entity must be at least min_len characters long.
            min_count (int, optional, default=0): The minimum frequency a named_entity must have in the corpus to be considered.
            format_ (str, optional, default=docterm-matrix): Format of the vectorization records.

                    Valid input values:
                    ["postings-list", "docterm-matrix"]

        Returns:
            Union[Dict[str, dict], Iterable[dict]]: If format_ == 'postings-list', returns a postings list; else a docterm matrix.

                    Example:
                    [format_ == "postings-list"]

                    {
                        "named_entity1":
                              {
                                "doc1": mi1,
                                ...
                              },
                        ...
                    }

                     Example:
                    [format_ == "docterm-matrix"]

                    [{
                        "doc": "doc1",
                        "bow":
                              {
                               named_entity1: mi1,
                               ...
                               }
                    },
                    ...
                    ]
        """


    @abc.abstractmethod
    def onehot(self, index: Iterable[dict], min_len: int = 0,
               min_count: int = 0, format_: str = 'docterm-matrix') -> Union[Dict[str, dict], Iterable[dict]]:
        """
        Calculates One-hot encoding.

        Args:
            index (Iterable[dict]): Records of an index produced during the tokenization stage (e.g. sentences, paragraphs, documents).

                    Example:
                    >>> sentence = next(sentences)
                    >>> print(sentence)
                    {
                    "sent_id_": 0,
                    "text_": "When We , though all unworthy , were called to succeed on the Apostolic Throne the meek Pius X [...]",
                    "lemma_": "when -PRON- , though all unworthy , be call to succeed on the Apostolic Throne the meek Pius X [ ... ]",
                    [...]
                    "named_entities_": [{"text": "Apostolic Throne", "start":12, "end":13}, {"text": "Pius X", "start": 16, "end":17}],
                    ...
                    }

            min_len (int, optional, default=0): The minimum length a named_entity must have to be considered. A named_entity must be at least min_len characters long.
            min_count (int, optional, default=0): The minimum frequency a named_entity must have in the corpus to be considered.
            format_ (str, optional, default=docterm-matrix): Format of the vectorization records.

                    Valid input values:
                    ["postings-list", "docterm-matrix"]

        Returns:
            Union[Dict[str, dict], Iterable[dict]]: If format_ == 'postings-list', returns a postings list; else a docterm matrix.

                    Example:
                    [format_ == "postings-list"]

                    {
                        "named_entity1":
                              {
                                "doc1": 0/1,
                                ...
                              },
                        ...
                    }

                     Example:
                    [format_ == "docterm-matrix"]

                    [{
                        "doc": "doc1",
                        "bow":
                              {
                               named_entity1: 0/1,
                               ...
                               }
                    },
                    ...
                    ]
        """

    @abc.abstractmethod
    def run(self,
            mode: str,
            index: Iterable[dict],
            normalize=True,
            min_len=0,
            min_count=0,
            metadata: Dict[str, dict] = None) -> Iterable[dict]:
        """
        Runs the vectorization grouping named_entities by self.ne_field and self.group_by_field.

        Args:
            mode (str): Vectorization mode. <br/>

                    Valid input values:
                    ["tf", "tfidf", "mi", "onehot"]

            index (Iterable[dict]): Records of an index produced during the tokenization stage (e.g. sentences, paragraphs, documents).

                    Example:
                    >>> sentence = next(sentences)
                    >>> print(sentence)
                    {
                    "sent_id_": 0,
                    "text_": "When We , though all unworthy , were called to succeed on the Apostolic Throne the meek Pius X [...]",
                    "lemma_": "when -PRON- , though all unworthy , be call to succeed on the Apostolic Throne the meek Pius X [ ... ]",
                    [...]
                    "named_entities_": [{"text": "Apostolic Throne", "start":12, "end":13}, {"text": "Pius X", "start": 16, "end":17}],
                    ...
                    }

            normalize (bool, optional, default=True): If True, normalizes TF for max(TF). <br/> This parameter is ignored if mode in ['tf', 'onehot', 'mi'].
            min_len (int, optional, default=0): The minimum length a named_entity must have to be considered. A named_entity must be at least min_len characters long.
            min_count (int, optional, default=0): The minimum frequency a named_entity must have in the corpus to be considered.
            metadata (Dict[str, dict], optional, default=None): A dict containing metadata for each different group_by_field value.  <br/> Each vector will be associated with its own metadata.

                    Example:

                    {
                     "group_by_field1": {
                                         metadata1: 'metadata1',
                                         metadata2: 'metadata2',
                                         ...
                                        },
                     "group_by_field2": {
                                          metadata1: 'metadata11',
                                          metadata2: 'metadata22',
                                         ...
                                        },
                    ...
                    }


        Returns:
            Iterable[dict]: Returns a docterm matrix.

                    Example:
                    >> v = NEVectorizer('doc_id', 'text')
                    >> record = v.run('tfidf', documents)
                    >> print(next(record))
                    {
                        "doc": "doc1",
                        "bow":
                              {
                               ne11: frequency1,
                               ...
                               }
                        metadata1: "metadata1",
                        metadata2: "metadata2",
                        ...
                    }
        """



language = 'norwegian-bokmål'
iso639 = 'nb'

spacy_model_sm = iso639 + '_core_news_sm'
spacy_model_md = iso639 + '_core_news_md'
spacy_model_lg = iso639 + '_core_news_lg'

sentiment_lexicon_class = None
pos_token_sentiment = None

negations = ['ikke', 'ikke', 'nei', 'ingen', 'verken', 'eller', 'aldri', 'ingen', 'intet', 'null', 'nullitet',
       'ingenting']


import subprocess
import setuptools
from setuptools.command.install import install

# -* dependencies *-
__install_requires__ = ['pandas', 'sklearn', 'matplotlib', 'numpy', 'spacy', 'nltk', 'tqdm',
                        'unicodedata2', 'gensim', 'configparser', 'vaderSentiment', 'unidecode',
                        'wordcloud']


# -* nltk.punkt, nltk.stopwords *-
class Install(install):
    def run(self):
        install.run(self)

        # install nltk punkt, stopwords
        for sub_module in ('punkt', 'stopwords'):
            cmd = ["python", "-m", "nltk.downloader", sub_module]
            with subprocess.Popen(cmd, stdout=subprocess.PIPE) as proc:
                print(proc.stdout.read())


# -* sample corpus: vatican publication *-
def doc_iterator(path: str) -> str:
    """Yields json documents."""
    import os

    for root, dirs, files in os.walk(path):
        for file in files:
            if file.endswith('.json'):
                yield os.path.join(root, file)

import pkg_resources
__sample_docs__ = list(doc_iterator(pkg_resources.resource_filename('compling', 'example-corpus/vatican-publications')))
__senti_lexicons__ = [pkg_resources.resource_filename('compling', 'senti-lexicons/sentix.csv'),
                      pkg_resources.resource_filename('compling', 'senti-lexicons/fannix.csv')]

# -* long description *-
with open("README.md", "r") as fh:
    __long_description__ = fh.read()

setuptools.setup(
    name="compling",
    version="0.0.23",
    author="Francesco Periti",
    author_email="peritifrancesco@gmail.com",
    description="Computational Linguistic",
    long_description=__long_description__,
    long_description_content_type="text/markdown",
    url="https://github.com/FrancescoPeriti/compling",
    packages=setuptools.find_packages(),
    package_data={'compling': ['config.ini', 'default_config.ini'] + __sample_docs__ + __senti_lexicons__},
    cmdclass={'install': Install},
    install_requires=__install_requires__,
    setup_requires=['nltk', 'spacy'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)

# python setup.py sdist bdist_wheel
# python -m twine upload --repository pypi dist/* --skip-existing

# dependency_links=[
# 'https://github.com/explosion/spacy-models/releases/download/en_core_web_sm-2.3.1/en_core_web_sm-2.3.1.tar.gz',
# 'https://github.com/explosion/spacy-models/releases/download/it_core_news_sm-2.3.0/it_core_news_sm-2.3.0.tar.gz',
# 'https://github.com/explosion/spacy-models/releases/download/es_core_news_sm-2.3.1/es_core_news_sm-2.3.1.tar.gz',
# 'https://github.com/explosion/spacy-models/releases/download/pt_core_news_sm-2.3.0/pt_core_news_sm-2.3.0.tar.gz',
# 'https://github.com/explosion/spacy-models/releases/download/fr_core_news_sm-2.3.0/fr_core_news_sm-2.3.0.tar.gz'
# 'https://github.com/explosion/spacy-models/releases/download/zh_core_web_sm-2.3.1/zh_core_web_sm-2.3.1.tar.gz',
# 'https://github.com/explosion/spacy-models/releases/download/da_core_news_sm-2.3.0/da_core_news_sm-2.3.0.tar.gz',
# 'https://github.com/explosion/spacy-models/releases/download/nl_core_news_sm-2.3.0/nl_core_news_sm-2.3.0.tar.gz',
# 'https://github.com/explosion/spacy-models/releases/download/de_core_news_sm-2.3.0/de_core_news_sm-2.3.0.tar.gz',
# 'https://github.com/explosion/spacy-models/releases/download/el_core_news_sm-2.3.0/el_core_news_sm-2.3.0.tar.gz',
# 'https://github.com/explosion/spacy-models/releases/download/ja_core_news_sm-2.3.2/ja_core_news_sm-2.3.2.tar.gz',
# 'https://github.com/explosion/spacy-models/releases/download/lt_core_news_sm-2.3.0/lt_core_news_sm-2.3.0.tar.gz',
# 'https://github.com/explosion/spacy-models/releases/download/nb_core_news_sm-2.3.0/nb_core_news_sm-2.3.0.tar.gz',
# 'https://github.com/explosion/spacy-models/releases/download/pl_core_news_sm-2.3.0/pl_core_news_sm-2.3.0.tar.gz',
# 'https://github.com/explosion/spacy-models/releases/download/ro_core_news_sm-2.3.1/ro_core_news_sm-2.3.1.tar.gz'
# ],



import csv
import pkg_resources
from compling.descriptor import paramtypecheck
from compling.lexicons.lexiconMeta import LexiconMeta

csv.field_size_limit(1000000)

# --* path fannix file *--
__fannix__ = pkg_resources.resource_filename('compling', 'senti-lexicons/fannix.csv')


class Fannix(LexiconMeta):
    """
    Sentiment Lexicon for Chinese Language.

    https://github.com/fannix/Chinese-Sentiment-Lexicon/
    """

    def __init__(self) -> None:
        """
        **\_\_init\_\_**: Creates a new Fannix Lexicon object for Chinese Sentiment Analysis.
        """

        super().__init__()

        # lexicon object
        self.__sentiment__ = dict()

        # read from file
        with open(__fannix__, encoding='utf-8', mode='r') as f:
            rows = csv.reader(f, delimiter=',')

            columns = next(rows)

            for row in rows:
                record = dict(zip(columns, row))
                lemma = record['lemma'].lower()
                self.__sentiment__[lemma] = record

    def polarity(self, token: str) -> dict:
        """
        Returns a float for sentiment strength based on the input text.
        Positive values are positive valence, negative value are negative valence, neutral value are neutral valence.'

        Args:
            token (str): The input token.

        Returns:
            dict: A positive, a negative and a neutral value.
        """

        polarities = dict({'neg': 0.0, 'pos': 0.0, 'obj': 0.0})

        # text processing
        is_negated = True if token.startswith('NOT_') else False
        if is_negated:
            token = token.replace("NOT_", "")
        token = token.lower()

        # polarities not available
        if token not in self.__sentiment__:
            return polarities

        # polarities available
        sentiment = self.__sentiment__[token]
        pos, neg, obj = [float(sentiment[p]) for p in polarities.keys()]

        # Inversion of polarity
        if is_negated:
            pos, neg = neg, pos

        # Positive - Negative
        polarities['pos'] += pos
        polarities['neg'] += neg

        # Neutral
        polarities['obj'] += obj

        return polarities



language = 'italian'
iso639 = 'it'

spacy_model_sm = iso639 + '_core_news_sm'
spacy_model_md = iso639 + '_core_news_md'
spacy_model_lg = iso639 + '_core_news_lg'

sentiment_lexicon_class = 'Sentix'

# Some lexicons provide a polarity of a token for each part of speech that token may have in the speech.
# If not None, only the polarities for the parts of speech in 'pos_token_sentiment' will be considered.
pos_token_sentiment = ['ADJ', 'ADV', 'VERB', 'NOUN', 'PROPN']

negations = ['nemmeno', 'neppure', 'neanche', 'nessuno', 'alcuno', 'veruno', 'niuno', 'niente', 'nulla', 'affatto',
             'mica', 'meno', 'mai più', 'figurati', 'figurarsi', 'mai', 'no', 'né', 'nè']



from typing import *
from compling.nlp import NLP
from compling.config import ConfigManager
import compling.lexicalAnalysis.tokenization.sentence_abc as abcsent
from compling.lexicalAnalysis.tokenization.impl.token import Token, f2att


class Sentence(abcsent.Sentence):
    def __init__(self, sent_id: int, token_id: int, sent: str, nlp: NLP, config: ConfigManager,
                 para_id: int, doc_id: int,
                 para_pos: int, doc_pos: int, metadata: dict) -> None:
        """
        Represents a generic sentence, which is a sequence of tokens in the input.
        A sentence is identified using a tokenizer, which breaks up the sequence of characters in the input into a sequence
        of tokens.

        Args:
           sent_id (int): Unique sentence identifier.
           token_id (int): Unique token identifier. The token_id of the first token in the sentence.
           sent (str): Text of the sentence.
           nlp (NLP): A Natural Language Processing object.
           config (ConfigManager): A Config manager object.
           para_id (int): Unique paragraph identifier. The id of the paragraph the sentence occurs in.
           doc_id (int): Unique document identifier. The id of the document the sentence occurs in.
           para_pos (int): Position index in the paragraph of the first token in the sentence.
           doc_pos (int): Position index in the document of the first token in the sentence.
           metadata (dict): Token metadata as a key-value pairs.

        Returns:
           None.
        """

        # unique identifiers
        self.token_id = token_id
        self.sent_id_ = sent_id

        # Natural Language Object
        self.nlp = nlp

        # sentence
        self.text_ = sent

        # position index
        self.para_pos = para_pos
        self.doc_pos = doc_pos

        # token list
        sentence = nlp.nlp_spacy(sent)
        # find negated token
        negation_mask = nlp.negated_tokens(sentence)
        self.token_list = [Token(self.next_token_id(), token, nlp, config,
                                 sent_id, para_id, doc_id,
                                 sent_pos, self.next_para_pos(), self.next_doc_pos(), negation_mask[sent_pos], metadata)
                           for sent_pos, token in enumerate(sentence)]

        # add some field only if the value in config.ini file is True
        for k, v in config.get_section(s='Sentence_record').items():
            if int(v):
                setattr(self, k + '_', getattr(self, k)())

        # add metadata for this sentence record
        self.add_metadata(metadata)

    @f2att
    def named_entities(self) -> List[Dict[str, Union[str, int]]]:
        """Returns named entities in the sentence."""

        named_entities = self.nlp.named_entities(self.text_)

        for ne_ in named_entities:
            ne_['sent_pos_start_'] = ne_.pop('start')
            ne_['sent_pos_end_'] = ne_.pop('end')
            ne_['text_'] = ne_.pop('text')

        # If n-grams are added in the text we don't want to find a named entities twice:
        # Examples:
        # Donald Trump __Donald_Trump__ made his speech in New York __New_York__
        # We don't want n-grams_entities:
        # [{'start': 0, 'end': 1, 'text': 'Donald Trump'}, {'start': 2, 'end': 2, 'text': 'Donald Trump'}, {'start': 7, 'end': 8, 'text': 'New York'}, {'start': 9, 'end': 9, 'text': 'New York'}]
        # We want ngram_entities_:
        # [{'start': 0, 'end': 1, 'text': 'Donald Trump'}, {'start': 7, 'end': 8, 'text': 'New York'}]

        if self.nlp.__ngrams_replaced__ == False:
            named_entities_ = list()
            for ne_ in named_entities:
                if ne_['sent_pos_start_'] == ne_['sent_pos_end_'] and len(ne_['text_'].split()) > 1:
                    continue
                else:
                    named_entities_.append(ne_)
            return named_entities_

        else:
            return named_entities

    def lang(self) -> str:
        """Returns the language of the sentence."""
        return self.nlp.iso639

    def next_para_pos(self) -> int:
        """Returns the next token position in the paragraph."""
        self.para_pos += 1
        return self.para_pos

    def next_doc_pos(self) -> int:
        """Returns the next token position in the document."""
        self.doc_pos += 1
        return self.doc_pos

    def next_token_id(self) -> int:
        """Returns the next token id."""
        self.token_id += 1
        return self.token_id

    def text(self) -> str:
        """Return the text of the sentence."""
        return self.text_

    def lower(self) -> str:
        """Return the lowercase version of the sentence text."""
        return self.text_.lower()

    @f2att
    def negations(self) -> str:
        """Return the text of the sentence where the text of negated tokens are preceded by the prefix 'NOT_'"""

        text = lambda token: 'NOT_' + token.text() if token.is_negated_ else token.text()
        return " ".join([text(token) for token in self.token_list])

    @f2att
    def sentiment(self) -> Dict[str, float]:
        """
        Returns the sentence polarities. <br/>
        Each polarity value is between 0 and 1. They are calculated as the average of the tokens polarities greater than 0. <br/>
        _p_ in _[0,1]_

        Returns:
            Dict[str, float]: Polarities type as keys, sentiment values as values.

                    Example:
                    sentence.text = "damn your friend, I hate her"

                    >> sentence.sentiment()
                    {
                     'pos' : 0.125,
                     'neg' : 0.775,
                     'obj' : 0.100
                    }
        """

        sentiment_values = {'pos': 0.0, 'neg': 0.0, 'obj': 0.0}
        count_gt0 = {'pos': 0, 'neg': 0, 'obj': 0.0}

        for token in self.token_list:
            for k in sentiment_values.keys():
                if token.sentiment()[k] > 0:
                    count_gt0[k] += 1
                sentiment_values[k] += token.sentiment()[k]

        for k in sentiment_values.keys():
            if count_gt0[k] > 0:
                if k in ("pos", "neg"):
                    sentiment_values[k] /= max(count_gt0["pos"], count_gt0["neg"])
                else:
                    sentiment_values[k] /= count_gt0[k]

        return sentiment_values

    @f2att
    def lemma(self) -> str:
        """Returns a version of the sentence text with tokens replace by their lemma."""

        return " ".join([token.lemma() for token in self.token_list])

    @f2att
    def stem(self) -> str:
        """Returns a version of the sentence text with tokens replace by their stem."""

        return " ".join([token.stem() for token in self.token_list])

    def add_metadata(self, metadata: dict) -> None:
        """Adds metadata as attributes of the sentence."""
        for k, v in metadata.items():
            if hasattr(self, k + '_'):
                continue
            setattr(self, k + '_', v)
        for token in self.token_list:
            token.add_metadata(metadata)

    def to_dict(self) -> dict:
        """Converts the sentence to a dict record."""
        return {k: v for k, v in self.__dict__.items() if k.endswith('_')}



import unidecode
import re
import spacy
import string
import unicodedata
from spacy.tokens.doc import Doc
from nltk.corpus import stopwords
from collections import defaultdict
from typing import *
from nltk.stem.snowball import SnowballStemmer
from compling.config import ConfigManager


class NLP:
    """
    Natural Language Processing (_NLP_) is a field of _Artificial Intelligence_ that gives the machines the ability to
    read, understand and derive meaning from human languages. It is a discipline that focuses on the interaction between
    data science and human language.
    """

    def __init__(self) -> None:

        # configuration
        self.config = ConfigManager()
        language = self.config.get(s='Corpus', k='language')

        # language module
        module = __import__('compling.languages.{}'.format(language))
        nlp_language = getattr(getattr(module, 'languages'), language)
        self.nlp_language = nlp_language

        # spacy model
        spacy_model_size = self.config.get(s='Corpus', k='spacy_model_size')
        spacy_model_available = {size: getattr(nlp_language, 'spacy_model_{}'.format(size))
                                 for size in ['sm', 'md', 'lg']}
        spacy_model = spacy_model_available[spacy_model_size]

        # natural language processing tool
        self.nlp_spacy = spacy.load(spacy_model)

        # language
        self.language = language

        # stemmer
        self.stemmer = SnowballStemmer(language)

        # lexicon
        if nlp_language.sentiment_lexicon_class is not None:
            module = __import__('compling.lexicons')
            self.lexicon = getattr(getattr(module, 'lexicons'), nlp_language.sentiment_lexicon_class)()
        else:
            # no lexicon available
            self.lexicon = None
            print('WARNING: No {} lexicon is available'.format(language))

    def is_stopword(self, token: str) -> bool:
        """
        Returns True if the input token is a stopword, False otherwise.

        Args:
            token (str): Text of the token that has to be checked.

        Returns:
            bool: True, if the token is a stopword, else False.
        """

        return token in self.stopwords_list()

    def token_sentiment(self, token: str) -> dict:
        """Returns the token polarities (positive, negative, neutral)."""

        return self.lexicon.polarity(token)

    @staticmethod
    def is_ngram(token: str, sep: str = '_', n: int = None) -> bool:
        """
        Returns True if the input token is a stopword, else False.

        Args:
            token (str): Text of the token that has to be checked.
            sep (str, optional, default=_): The character separator that splits the n-gram into tokens.
            n (int, optional, default=None): If not None, checks if the input token is a n-gram of size n.

        Returns:
            bool: True, if the token is a n-gram, else False.
        """
        return token.count(sep) > 0 if n is None else token.count(sep) >= n - 1

    @staticmethod
    def is_too_short(text: Union[str, list], min_len: int) -> bool:
        """
        If text is instance of str, returns True if text is less than 'min_len' characters long else False.
        If text is isntance of list, return True if text is less than 'min_len' tokens long else False.

        Args:
            text (Union[str, List[str]]): The input text that has to be checked.
            min_len (int, optional, default=0): Threshold: minimun length

        Returns:
            bool: True, if text is too short, else False.
        """
        return len(text) <= min_len

    @staticmethod
    def is_lower(text: str) -> bool:
        """
        Returns True if text is a lowercase string, False otherwise.

        Args:
            text (str): The input text.

        Returns:
            bool: True, if the text is a lowercase string, False otherwise.
        """

        return text.islower()

    @staticmethod
    def is_upper(text: str) -> bool:
        """
        Returns True if text is a uppercase string, False otherwise.

        Args:
            text (str): The input text.

        Returns:
            bool: True, if the text is a uppercase string, False otherwise.
        """
        return text.isupper()

    @staticmethod
    def is_capitalize(text: str) -> bool:
        """
        Returns True if text is a capitalize string, False otherwise.

        Args:
            text (str): The input text.

        Returns:
            bool: True, if the text is a capitalize string, False otherwise.
        """

        if len(text) < 2:
            return False

        return True if text[0].isupper() and text[1:].islower() else False

    @staticmethod
    def is_title(text: str) -> bool:
        """
        Returns True if text is a title string, False otherwise.

        Args:
            text (str): The input text.

        Returns:
            bool: True, if the text is a title string, False otherwise.
        """
        return text.istitle()

    @staticmethod
    def is_digit(token: str) -> bool:
        """
        Returns True if token is a digit string, False otherwise.

        Args:
            token (str): Text of the token that has to be checked.

        Returns:
            bool: True, if the token is a digit string, False otherwise.
        """
        return token.isdigit()

    @staticmethod
    def is_punct(token: str) -> bool:
        """
        Returns True if token is a punctuation character, False otherwise.

        Args:
            token (str): Text of the token that has to be checked.

        Returns:
            bool: True, if the token is a punctuation character, False otherwise.
        """
        return token in string.punctuation

    @staticmethod
    def is_space(token: str) -> bool:
        """
        Returns True if token is a space character, False otherwise.

        Args:
            token (str): Text of the token that has to be checked.

        Returns:
            bool: True, if the token is a space character, False otherwise.
        """
        return token in string.whitespace

    @staticmethod
    def strip_punctuation(text: str, keep: Iterable[str] = None, sc: Iterable[str] = None) -> str:
        """
        Strips punctuation characters from a input text.

        Args:
            text (str): The input text.
            keep (Tuple[str, ...], optional, default=None): Punctuation characters to keep: they won't be stripped.
            sc (Tuple[str, ...], optional, default=None): Additional Special characters to be stripped.

        Returns:
            str: The text without punctuation characters.
        """

        punctuation = string.punctuation
        if keep is not None and len(keep) > 0:
            punctuation = "".join([i for i in punctuation if i not in keep])
        if sc is not None and len(sc) > 0:
            punctuation += "".join(sc)

        if keep is not None:
            for k in keep:
                text = text.replace(k, ' ' + k + ' ')

        return text.translate(str.maketrans(dict.fromkeys(punctuation, ' '))).strip()

    @staticmethod
    def strip_accents(text: str) -> str:
        """
        Strips accents from a input text.

        Args:
            text (str): The input text.

        Returns:
            str: The text without accents characters.
        """
        return ''.join(c for c in unicodedata.normalize('NFD', text) if unicodedata.category(c) != 'Mn').strip()

    def strip_stopwords(self, text: str, min_len: int = 0) -> str:
        """
        Strips accents from a input text.

        Args:
            text (str): The input text.
            min_len (int, optional, default=0): Tokens shorter than min_len are stripped as stopwords.

        Returns:
            str: The text without accents characters.
        """
        return " ".join(
            [token for token in text.split() if not self.is_stopword(token) and len(token) >= min_len]).lower()

    @staticmethod
    def lower(text: str) -> str:
        """
        Returns a copy of the string converted to lowercase.

        Args:
            text (str): The input text.

        Returns:
            str: A copy of the string converted to lowercase.
        """
        return text.lower()

    @staticmethod
    def upper(text: str) -> str:
        """
        Returns a copy of the string converted to uppercase.

        Args:
            text (str): The input text.

        Returns:
            str: A copy of the string converted to uppercase.
        """
        return text.upper()

    @staticmethod
    def capitalize(text: str) -> str:
        """
        Returns a copy of the string converted to capitalize.

        Args:
            text (str): The input text.

        Returns:
            str: A copy of the string converted to capitalize.
        """
        return text.capitalize()

    @staticmethod
    def title(text: str) -> str:
        """
        Returns a copy of the string converted to title.

        Args:
            text (str): The input text.

        Returns:
            str: A copy of the string converted to title.
        """
        return text.title()

    def stem(self, token: str) -> str:
        """
        Strips affixes from the token and returns the stem.

        Args:
            token (str): The text of the input token.

        Returns:
            str: Returns the stem of the token.
        """
        return self.stemmer.stem(token)

    def split_skipgram(self, text: str, ws: int = 3) -> List[str]:
        """
        Returns a list of tokens as union of the tokens of each skipgrams pair.

        Args:
            text (str): The input text.
            ws (int, optional, default=3): Window size: size of sampling windows. <br/> The window of a word w_i will be [i-window_size, i+window_size+1]

        Returns:
            List[str]: A list of tokens as union of the tokens of each skipgrams pair.

                    Example:
                    >> nlp = NLP()
                    >> nlp.split_skipgram(text="My mistress' eyes are nothing like the sun", ws=2)
        """

        tokens = list()
        for skipgram in self.skipgrams(text, ws):
            tokens.extend(skipgram)
        return tokens

    def stopwords_list(self, include: Iterable[str] = None) -> List[str]:
        """
        Returns a stopwords list.

        Args:
            include (List[str], optional, default=None): Include a list of arbitrary stopwords.

        Returns:
            List[str]: a list of stopwords.
        """

        if not hasattr(self, 'stopwords'):
            self.stopwords = stopwords.words(self.language)
            if include is not None:
                self.stopwords.extend([s.lower() for s in include])

        return self.stopwords

    def ngrams(self, text: str, n: int, pos: Iterable[str] = ("PROPN", "VERB", "NOUN", "ADJ"),
               threshold: int = 50) -> Dict[str, int]:
        """
        Returns the most frequent n-grams in the text.

        Args:
            text (str): Input text.
            n (int): The size of n-grams: number of sequential strings that make up n-grams.
            pos (Tuple[str, ...], optional, default=PROPN,VERB,NOUN,ADJ): Part of speech of the first and the last token that make up n-grams. <br/> Filters the most informative n-grams. <br/> If None, the filter will be ignored.
            threshold (int, optional, default=50): Filters n-grams that have a frequency greater than threshold.

        Returns:
             Dict[str, int]: N-grams as keys, frequencies as values.
        """

        # Ngrams dict {ngram: frequency}
        ngram_frequencies = defaultdict(lambda: 1)

        # Scan of each source using a Sliding Window
        # Index Sliding Window
        c = 0

        # From text source to token list
        tokens = text.split()

        # Shift the Sliding Window until it covers the last token
        while c < len(tokens) - (n - 1):
            c += 1

            # First and last word can't be stopwords
            if self.is_stopword(tokens[c - 1]) or \
                    self.is_stopword(tokens[c + (n - 1) - 1]):
                continue

            ngram = tuple([tokens[c + i - 1] for i in range(0, n)])
            ngram_frequencies[ngram] = ngram_frequencies[ngram] + 1

        # Select the n grams you're interested in
        ngram_frequencies = {ngram: frequency for ngram, frequency in ngram_frequencies.items() if
                             frequency > threshold}

        if pos is None or len(pos) == 0:
            return ngram_frequencies

        # Ngram final list
        result = defaultdict(int)

        # First and last word must be in pos
        for ngram in ngram_frequencies:
            ngram_ = self.nlp_spacy(" ".join(ngram))
            if ngram_[0].pos_ in pos and ngram_[-1].pos_ in pos:
                ngram_ = " ".join([token.text for token in ngram_])
                result[ngram_] = ngram_frequencies[ngram]

        return result

    @staticmethod
    def skipgrams(text: str, ws: int = 3) -> List[Tuple[str, str]]:
        """
        Generates skipgram word pairs.

        Args:
            text (str): The input text.
            ws (int, optional, default=3): Window size: size of sampling windows. <br/> The window of a word w_i will be [i-window_size, i+window_size+1]

        Returns:
            List[Tuple[str, str]]: A list of tuple (skipgram pairs).
        """

        skipgrams = []
        tokens = text.split()

        ws += 1
        for i in range(0, len(tokens) - 1):
            for j in range(1, ws + 1):
                if i != len(tokens) - j and i + j < len(tokens):
                    skipgrams.append((tokens[i], tokens[i + j]))

        return skipgrams

    def named_entities(self, text: str) -> List[Dict[str, Union[str, int]]]:
        """
        Returns a list of named entities in text.

        Args:
            text (str): The input text.

        Returns:
            List[Dict[str, Union[str, int]]]: A list of named entities. Each named entity is a dict with three keys:

                * _text_: the named entity
                * _start_: the position of the first token of the Named Entities in the text.
                * _end_: the position of the last token of the Named Entities in the text.
        """

        # cler text
        text = unidecode.unidecode(text)
        #text = self.strip_punctuation(text)

        if len(text) < 2:
            return []

        # find Named Entities by regex
        #"(?:(?<=^)|(?<=[^.]))\s+([A-Z]\w+)"
        named_entities = [{'start': m.start(0), 'end': m.end(0), 'text': m.group(0)}
                          for m in re.finditer("([A-Z]+[a-z]*)", text[0].capitalize() + text[1:], re.UNICODE)
                          if not self.is_stopword(m.group(0).lower())]

        text_spans = self.spans(text)

        # One or less named entities
        if len(named_entities) == 0:
            return list()
        elif len(named_entities) == 1:
            # pos start, pos end for each named entities
            for k, v in text_spans.items():
                if named_entities[0]['start'] in range(*k):
                    named_entities[0]['start'] = v
                    break

            for k, v in text_spans.items():
                if (named_entities[0]['end'] - 1) in range(*k):
                    named_entities[0]['end'] = v
                    break
            return [named_entities[0]]

        # Link consecutive named entities
        result = [named_entities[0]]
        for named_entity in named_entities[1:]:
            if result[-1]['end'] == named_entity['start'] - 1:
                result[-1]['text'] += " "+named_entity['text']
                result[-1]['end'] = named_entity['end']
            else:
                result.append(named_entity)

        # pos start, pos end for each named entities
        for ne in result:
            for k, v in text_spans.items():
                if ne['start'] in range(*k):
                    ne['start'] = v
                    break

            for k, v in text_spans.items():
                if (ne['end'] - 1) in range(*k):
                    ne['end'] = v
                    break
        return result

    def spacy_named_entities(self, text: str) -> List[Dict[str, Union[str, int]]]:
        """
        Returns a list of named entities in text recognized by spacy.

        Args:
            text (str): The input text.

        Returns:
            List[Dict[str, Union[str, int]]]: A list of named entities. Each named entity is a dict with three keys:

                * _text_: the named entity
                * _start_: the position of the first token of the Named Entities in the text.
                * _end_: the position of the last token of the Named Entities in the text.
        """

        named_entities = []
        for ent in self.nlp_spacy(text).ents:
            named_entity = dict()
            named_entity['text'] = ent.text
            named_entity['start'] = -1
            named_entity['end'] = -1

            character_index = 0
            for token_index, token in enumerate(text.split()):
                for _ in token:
                    if ent.start_char == character_index:
                        named_entity['start'] = token_index
                    if ent.end_char == character_index + 1:
                        named_entity['end'] = token_index
                    character_index += 1

                # whitespace character
                character_index += 1

                if named_entity['start'] >= 0 and named_entity['end'] >= 0:
                    break

            named_entities.append(named_entity)

    @staticmethod
    def spans(text: str) -> Dict[Tuple[int, int], int]:
        """
        Returns the starting and ending character index for each token in the text.

        Args:
            text (str): The input text.

        Returns:
            Dict[Tuple[int, int], int]: A dict containing the the starting and ending character index for each token in the text as keys;
            and the token index for each token in text as values.
        """

        result = dict()
        point = 0  # Where we're in the text.

        for t, fragment in enumerate(text.split()):
            found_start = text.index(fragment, point)
            found_end = found_start + len(fragment)
            result[(found_start, found_end)] = t
            point = found_end

        return result

    def negated_tokens(self, text: Union[str, Doc]) -> List[bool]:
        """Returns a mask that specifies which tokens are negated in the text.

        Args:
            text (Union[str, spacy.tokens.doc.Doc]): The input text that has to be analyzed.

        Returns:
            bool: A mask that specifies which tokens are negated in the text.
        """

        tokens = text if type(text) is Doc else self.nlp_spacy(text)
        mask, negated = [], []

        # All children tokens linked to a parent token by a negation relationship
        for token in tokens:
            if token.dep_ == "neg" or token.text in self.nlp_language.negations:
                negated += [x for x in token.head.children] + [token.head]

        for token in tokens:

            # All children tokens linked to a parent token by a negation relationship and
            # All tokens attached to a modifying adjective or adverb
            if (token in negated or (token.dep_ in ["amod", "advmod"] and token.head in negated)) and \
                    token.text not in self.nlp_language.negations:
                mask.append(True)
                continue

            mask.append(False)
        return mask


import abc
from compling.nlptoolkit import NLP
import gensim
from typing import *


class Embeddings(metaclass=abc.ABCMeta):
    """
    A word embedding is a learned representation for text where words that have the same meaning have a similar representation.

    Word embeddings are a class of techniques where individual words are represented as real-valued vectors in a predefined vector space. Each word is mapped to one vector and the vector values are learned in a way that resembles a neural network, and hence the technique is often lumped into the field of deep learning.

    Key to the approach is the idea of using a dense distributed representation for each word.

    Each word is represented by a real-valued vector, often tens or hundreds of dimensions. This is contrasted to the thousands or millions of dimensions required for sparse word representations, such as a one-hot encoding.
    """

    def __init__(self, index: Iterable[dict], module: gensim.models, output:str=None, text_field='text',
                 skipgram_ws: int = None, tag=False, id_field:Union[str, List[str]]=None) -> None:
        """
        **\_\_init\_\_**: Creates a new Embeddings instance.

        Args:
           index (Iterable[dict]): Records of an index produced by tokenization.
           module (gensim.models): Gensim module.

                   Valid input:
                   - gensim.models.Word2Vec;
                   - gensim.models.fasttext.FastText;
                   - gensim.models.doc2vec.Doc2Vec

           output (str, optional, default=None): If not None, stores the model with the name in output.
           text_field (str, optional, default=text): The text field of the records.
           skipgram_ws (int, optional, default=None): If not None, text inside records was skipgram_splitted with skipgram_ws=skipgram_ws. <br/> Window size: size of sampling windows. <br/> The window of a word _w_i_ will be _[i-window_size, i+window_size+1]_ <br/> See: nlp.NLP.split_skipgram
           tag (bool, optional, default=True): If True, Tag each record with its id_field.
           id_field (Union[str, List[str]], optional, default=None): Id fields conteining tag for each records.
        """

        if id_field is not None:
            if isinstance(id_field, str):
                id_field = [id_field]

            id_field = " + ".join([i+'_' for i in id_field])

        self.id_field = id_field

        text_field += '_'

        # natural language processing object
        self.nlp = NLP()

        self.module = module
        self.model = None
        self.output = output

        if tag:
            self.data = [gensim.models.doc2vec.TaggedDocument(words=record[text_field].split(),
                                                              tags=[str(record[self.id_field])])
                         if not skipgram_ws else
                         gensim.models.doc2vec.TaggedDocument(
                             words=self.self.nlp.split_skipgram(record[text_field], ws=skipgram_ws),
                             tags=[str(record[self.id_field])])
                         for record in index]

        else:
            self.data = [record[text_field].split() if not skipgram_ws else
                         self.nlp.split_skipgram(record[text_field], ws=skipgram_ws)
                         for record in index]

    def __getattr__(self, name):
        value = super(Embeddings, self).__getattr__(self, name)
        if name == 'model' and value is None:
            return self.load()
        else:
            return value

    def load(self):
        """Load a previously saved Embeddings model from file."""

        self.model = self.module.load(self.path)
        return self.model

    def save(self):
        """Stores the model. This saved model can be loaded again using load(), which supports
        getting vectors for vocabulary words."""

        self.model.save(self.output)

    @abc.abstractmethod
    def most_similar_threshold(self, positive: Union[list, str] = None,
                               negative: Union[list, str] = None,
                               threshold: float = 0.75) -> Dict[str, float]:
        """
        Finds the most similar vectors, whose cosine similarity is lower than threshold.
        Positive words/doc_ids contribute positively towards the similarity, negative words/doc_ids negatively.
        This method computes cosine similarity between a simple mean of the projection weight vectors of the
        given words/doc_ids and the vectors for each word/doc_id in the model.

        Args:
            positive (Union[list, str]): List of words/doc_ids that contribute positively.
            negative (Union[list, str]): List of words/doc_ids that contribute negatively.
            threshold (float, optional, default=0.75): Cosine similarity threshold.

        Returns:
            Dict[str, float]: Words/doc_ids as keys, cosine similarity values as values.
        """

    @abc.abstractmethod
    def most_similar(self, positive: Union[list, str] = None,
                     negative: Union[list, str] = None,
                     topn: int = 10) -> Dict[str, float]:
        """
        Finds the top-N most similar words/doc_ids.
        Positive words/doc_ids contribute positively towards the similarity, negative words/doc_ids negatively.
        This method computes cosine similarity between a simple mean of the projection weight vectors of the
        given words/doc_ids and the vectors for each word/doc_id in the model.

        Args:
            positive (Union[list, str]): List of words/doc_ids that contribute positively.
            negative (Union[list, str]): List of words/doc_ids that contribute negatively.
            topn (int, optional, default=10): Number of top-N similar words to return.

        Returns:
            Dict[str, float]: Words/doc_ids as keys, cosine similarity values as values.
        """

    @abc.abstractmethod
    def similarity(self, id1: str, id2: str) -> float:
        """
        Computes cosine similarity between two words/doc_ids.

        Args:
            id1 (str): First word/doc_id.
            id2 (str): First word/doc_id.

        Returns:
            float: Cosine similarity between two vectors.
        """

    def analogy(self, positive: List[str], negative: List[str], topn=10) -> Dict[str, float]:
        """
        Inferes an analogy.
        This method is the same of most_similar but forces to have positive and negative List[str] as input.

            Example:
            >> analogy(positive=[king, woman], negative=[man]) -> {queen: 0.99, ...}

        Positive words/doc_ids contribute positively towards the similarity, negative words/doc_ids negatively.
        This method computes cosine similarity between a simple mean of the projection weight vectors of the
        given words/doc_ids and the vectors for each word/doc_id in the model.

        Args:
            positive (Union[list, str]): List of words/doc_ids that contribute positively.
            negative (Union[list, str]): List of words/doc_ids that contribute negatively.
            topn (int, optional, default=10): Number of top-N similar words to return.

        Returns:
            Dict[str, float]: Words/doc_ids as keys, cosine similarity values as values.
        """
        return self.most_similar(positive, negative, topn)

    def run(self, **kwargs) -> None:
        """Trains a Neural Network and builds a Embeddings Model."""

        param_dict = {k: v for k, v in locals()['kwargs'].items() if k not in ('self', 'epochs') and v is not None}
        model = self.module(**param_dict)

        model.build_vocab(self.data)

        model.train(self.data,
                    epochs=locals()['kwargs']['epochs'],
                    total_examples=model.corpus_count,
                    total_words=model.corpus_total_words)
        self.model = model

        if self.output is not None:
            self.save()



from typing import *
from tqdm import tqdm
from datetime import datetime
from collections import defaultdict
from nltk.tokenize import sent_tokenize
from compling.lexicalAnalysis.tokenization import tokenization_abc
from collections.abc import Iterable as abciterable
from compling.lexicalAnalysis.tokenization.impl.sentence import Sentence
from compling.lexicalAnalysis.tokenization.impl.paragraph import Paragraph
from compling.lexicalAnalysis.tokenization.impl.document import Document
#from compling.descriptor import paramtypecheck, paramtypechecktarget


class Tokenizer(tokenization_abc.Tokenization):
    """
    Tokenization converts input text to streams of tokens, where each token is a separate word, punctuation sign,
    number/amount, date, etc.

    A Tokenizer object converts the corpus documents into a stream of:

       * _tokens_: tokens occurring in those documents. Each token is characterized by:
          * _token_id_: unique token identifier;
          * _sent_id_: unique sentence identifier. The id of the sentence the token occurs in;
          * _para_id_: unique paragraph identifier. The id of the paragraph the token occurs in;
          * _doc_id_: unique document identifier. The id of the document the token occurs in;
          * _text_: the text of the token;
          * a large variety of _optional meta-information_ (e.g. PoS tag, dep tag, lemma, stem, ...);
       * _sentences_ : sentences occurring in those documents. Each sentence is characterized by:
          * _sent_id_: unique sentence identifier;
          * _para_id_: unique paragraph identifier. The id of the paragraph the sentence occurs in;
          * _doc_id_: unique document identifier. The id of the document the sentence occurs in;
          * _text_: the text of the sentence;
          * a large variety of _optional meta-information_ (e.g.lemma, stem, ...);
       * _paragraphs_: sentences occurring in those documents. Each paragraph is characterized by:
          * _para_id_: unique paragraph identifier;
          * _doc_id_: unique document identifier. The id of the document the paragraph occurs in;
          * _text_: the text of the paragraph;
          * a large variety of _optional meta-information_ (e.g.lemma, stem, ...);
       * _documents_: Each document is characterized by:
          * _doc_id_: unique document identifier;
          * _text_: the text of the document;
          * a large variety of _optional meta-information_ (e.g.lemma, stem, ...);

    A Tokenizer object is also able to retrieve frequent n-grams to be considered as unique tokens.

    For each record (token, sentence, paragraph, document) are stored some metadata.
    You can edit the config.ini file to change those stored by default.

    In order to run tokenization you need to provide a Iterable[dict], where each document is a dict
    and has a key where it stores the text of the document. By default the text key is 'text', you can change it
    editing the config.ini file.

    For each document in your corpus, all key/value data (except for text key) are added as metadata to
    that document records. (e.g. title, author, ...).
    """

    def __init__(self, domain_stopwords: Iterable[str] = None) -> None:
        """
        **\_\_init\_\_**: Creates a new Tokenizer object.

        Args:
            domain_stopwords (List[str], optional, default=None): You can provide a list of arbitrary stopwords specific to your corpus domain.
        """

        if domain_stopwords is not None and not all([isinstance(s, str) for s in domain_stopwords]):
            raise TypeError('domain_stopwords must be Iterable[str], not {}'.format(type(domain_stopwords).__name__))

        # super __init__
        super(Tokenizer, self).__init__()

        # add some stopwords to nlp list.
        self.nlp.stopwords_list(include=domain_stopwords)
        self.nlp.__ngrams_replaced__ = True

        self.corpus_ngrams = None

    #@paramtypechecktarget(params=['corpus_threshold', 'doc_threshold', 'len_gram', 'replace'])
    def ngrams2tokens(self,
                      n: Union[int, Iterable[int]],
                      docs_in: Iterable[Dict[str, str]],
                      docs_out: Iterable[Dict[str, str]] = None,
                      pos: List[str] = ("PROPN", "VERB", "NOUN", "ADJ"),
                      corpus_threshold: int = 50,
                      doc_threshold: int = 0,
                      len_gram: int = 3,
                      include: List[str] = None,
                      replace: bool = False) -> Iterable[Dict[int, str]]:
        """
        Calculates frequent n-grams in the corpus. They will be considered as unique tokens during the Tokenization task. <br/>
        Frequent n-grams will be calculated based on the documents in the input stream (docs_in). <br/>
        Frequent n-grams will be considered as unique tokens in the output stream documents (docs_out).

        Args:
           docs_in (Iterable[Dict[str, str]]): Stream of input json documents as dicts. <br/> Each document must have a key where the text of the document is stored. Frequent n-grams will be calculated by scrolling this stream.
           docs_out (Iterable[Dict[str, str]]): Stream of output json documents as dicts. <br/> Each document must have a key where the text of the document is stored. Frequent n-grams will be considered as unique tokens for each document in the stream. Each document will be yielded as output of this function.
           n (Union[int, Iterable[int]]): If int, the size of n-grams. (e.g. n=2, bigrams) <br/> If Iterable[int], the sizes of n-grams. (e.g. n=[2,3,5], bigrams, trigrams, fivegrams)
           pos (Tuple[str, ...], optional): default ('PROPN', 'VERB', 'NOUN', 'ADJ'). <br/> Part of speech of the first and the last token that make up n-grams. <br/> Filters the most informative n-grams. <br/> If None, the filter will be ignored.

                   Example:
                   pos = ("PROPN", "VERB", "NOUN", "ADJ")

                   These n-grams are IGNORED:
                   - "of a": of [ADP], a [DET]
                   - "at home":  at [ADP], home [NOUN]
                   - "as much then": as [ADP], as [ADP]
                   - "a computer scientist": a [DET], scientist [NOUN]
                   - "of natural phenomena": of [ADP], phenomena [NOUN]
                   - ...

                   These n-grams are CONSIDERED:
                   * "mother Earth": mother [NOUN], Earth [PROPN]
                   * "John likes": John [PROPN], likes [VERB]
                   * "computer scientist": computer [NOUN], scientist [NOUN]
                   * "Galilean scientific method": Galilean [ADJ], method [NOUN]
                   * "understanding of natural phenomena": understanding [NOUN], phenomena [NOUN]
                   ...

           corpus_threshold (int, optional, default=50): Filters n-grams that have a corpus frequency greater than corpus_threshold.
           doc_threshold (int, optional, default=0): Filters n-grams that have frequency greater than doc_threshold. <br/> The frequency of an n-gram in the corpus is the sum of the frequency of that n-gram in documents where the ngram occurs at least doc_thresold times.
           len_gram (int, optional, default=3): The size of the first and the last token that make up n-grams must be at least 'len_gram'.

                   Example:
                   len_gram = 5

                   These n-grams are IGNORED:
                   - "John likes": John [4], likes [5]
                   - "New York":  New [3], York [3]
                   - ...

                   These n-grams are CONSIDERED:
                   * "mother Earth": mother [6], Earth [5]
                   * "computer scientist": computer [8], scientist [9]
                   ...

           include (Tuple[str, ...], optional, default=None): Include a list of arbitrary n-grams.
           replace (bool, optional, default=True): If True, replaces a n-gram with its tokens separated by '_'. <br/> Else, concatenates a new token, made merging the n-gram tokens with '_', to the n-gram.

                   Example:

                    - replace = True:
                      "New York is the most populous city in the United States."
                       ->
                      "New_York is the most populous city in the United_States."

                    - replace = False:
                      "New York is the most populous city in the United States."
                       ->
                      "New York New_York is the most populous city in the United States United_States."

        Returns:
            Iterable[Dict[str, str]]: Stream of json documents (docs_out) with frequent n-grams considered as unique token.


            It will be setted _self.corpus_ngram_ (Dict[str, int]): N-grams as keys, frequencies as values. <br/> The frequencies of the n-grams included arbitrarily are set to 0.

                    Example:
                    include = ("Giorgio Napolitano")

                     {
                        "New York": 51,
                        "Giorgio Napolitano": 0,
                        ...
                        "Donald Trump": 402
                     }
        """

        # list of sizes of n-grams
        if isinstance(n, abciterable):
            sizes = n
        elif isinstance(n, int):
            sizes = [n]
        else:
            raise TypeError('n must be typing.Union[int, Iterable[int], not {}'.format(type(n).__name__))

        for k, p in {'pos': pos, 'include': include}.items():
            if p is not None and not all([isinstance(i, str) for i in p]):
                raise TypeError('{} must be typing.Iterable[str], not {}'.format(k, type(p).__name__))

        ngram_frequencies = defaultdict(int)

        # for each corpus doc
        for doc in tqdm(docs_in, desc='N-gram Retrieval in progress...', position=0, leave=True):
            text = doc[self.text_key]
            for n in sizes:
                for ngram, frequency in self.nlp.ngrams(text, n, pos=tuple(), threshold=doc_threshold).items():
                    ngram_frequencies[ngram] += frequency

        # Select the n grams you're interested in
        ngram_frequencies = {ngram: frequency for ngram, frequency in ngram_frequencies.items() if
                             frequency >= corpus_threshold}

        # Ngram final list
        result = defaultdict(int)

        # First and last word must be in pos and at least len_gram sized
        for ngram in ngram_frequencies:
            ngram_ = self.nlp.nlp_spacy(" ".join(ngram))
            if ngram_[0].pos_ in pos and ngram_[-1].pos_ in pos and \
                    len(ngram_[0].text) >= len_gram and len(ngram_[-1].text) >= len_gram:
                ngram_ = " ".join([token.text for token in ngram_])
                result[ngram_] = ngram_frequencies[ngram]

        # add arbitrary n-grams
        if include is not None:
            for ngram in include:
                if ngram not in result:
                    result[ngram] = 0

        # sorted by len: first we replace the longer n-grams, then shorter ones.
        self.corpus_ngrams = dict(sorted(result.items(), key=lambda x: len(x[0]), reverse=True))

        # updates doc: replaces n-grams with tokens
        if docs_out is None:
            return

        if replace:
            self.nlp.__ngrams_replaced__ = replace

        for doc in docs_out:
            for ngram in self.corpus_ngrams:

                ngram = ngram.replace('.', '\u002E')

                if replace:
                    doc[self.text_key] = doc[self.text_key].replace(ngram, "__" + ngram.replace(' ', '_') + "__")
                else:
                    doc[self.text_key] = doc[self.text_key].replace(ngram,
                                                                    ngram + ' __' + ngram.replace(' ', '_') + "__")

            yield doc

    #@paramtypechecktarget(params=['doc_id', 'token_id', 'sent_id', 'para_id', 'para_size', 'index_doc',
    #                              'index_sent', 'index_para'])
    def run(self,
            docs: Iterable[Dict[str, str]],
            doc_id: int = 0,
            token_id: int = 0,
            sent_id: int = 0,
            para_id: int = 0,
            para_size: int = 3,
            index_doc: bool = True,
            index_sent: bool = True,
            index_para: bool = True) -> Iterable[Dict[str, dict]]:
        """
        Runs the tokenization for each document of the corpus.

        Args:
            docs (Iterable[Dict[str, str]]): Stream of json documents as dicts. <br/> Each document must have a key where the text of the document is stored. Each document will be tokenized.
            doc_id (int, optional, default=0): Unique document identifier. <br/> Numeric id of documents. <br/> A progressive id will be assigned to the documents starting from this number.
            token_id (int, optional, default=0): Unique token identifier. <br/> Numeric id of tokens. <br/> A progressive id will be assigned to the tokens starting from this number.
            sent_id (int, optional, default=0): Unique sentence identifier. <br/> Numeric id of sentences. <br/> A progressive id will be assigned to the sentences starting from this number.
            para_id (int, optional, default=0): Unique paragraph identifier. <br/> Numeric id of paragraphs. <br/> A progressive id will be assigned to the paragraphs starting from this number.
            para_size (int, optional, default=3): The paragraph size: the number of sentences that makes up a paragraph.
            index_doc (bool, optional, default=True): If True, returns the records of the documents index as values of the key 'documents'.
            index_sent (bool, optional, default=True): If True, returns the records of the sentences index as values of the key 'sentences'.
            index_para (bool, optional, default=True): If True, returns the records of the paragraphs index as values of the key 'paragraphs'.

        Returns:
            Iterable[Dict[str, dict]]: Stream of tokens, sentences, paragraphs, documents records.

                    Example:
                    [index_sent == True, index_para == False, index_doc == True]

                    >>> t = Tokenizer(domain_stopwords=my_domain_stopwords)
                    >>> records = t.run(docs=my_docs):
                    >>> next(records).keys()
                    dict_keys(["tokens", "sentences", "documents"])


                    Example:
                    >>> record = next(records)
                    >>> tokens = record["tokens"]
                    >>> sentences = record["sentences"]
                    >>> print(tokens)
                    [
                        {
                        "token_id_": 0,
                        "sent_id_": 0,
                        "para_id_": 0,
                        "doc_id_": 0,
                        "text_": "When",
                        "sent_pos_": 0,
                        "para_pos_": 0,
                        "doc_pos_": 0,
                        "lower_": "when",
                        ...
                        },
                        ...
                    ]
                    >>> print(sentences)
                    [
                        {
                         "sent_id_": 0,
                         "text_": "When We , though all unworthy , were called to succeed on the Apostolic Throne the meek Pius X , whose life of holiness and well doing was cut short by grief -atti_degli_apostoli- the fratricidal struggle that had just burst forth in Europe , We , too , on turning a fearful glance on the blood stained battlefields , felt the anguish of a father , who sees his homestead devastated and in ruins before the fury of the hurricane .",
                         "named_entities_": ["Apostolic Throne", "Pius X", "Europe"],
                        ...
                        }
                    ]


                Check the config.ini file for records information.
        """

        # list of sentences making up a paragraph
        # list of paragraphs making up a document
        sent_list, para_list = list(), list()

        # SIDE EFFECT: Sentence will increment token_id by one, so that it can start from token_id and not from token_id+1 .
        token_id = token_id - 1

        # tokenization of each doc
        for doc in tqdm(docs, desc='Tokenization in progress...', position=0, leave=True):

            text = doc[self.text_key]

            metadata = dict()
            for k, v in doc.items():
                if k == self.text_key:
                    continue
                try:
                    # if it's a date
                    v = datetime.strptime(v, self.date_format)
                except:
                    pass
                metadata[k] = v

            # token position inside the paragraphs/documents: -1, Sentence will increment it, so that it can start from 0.
            para_pos, doc_pos = -1, -1

            # tokenization of each sentence
            for sent in sent_tokenize(text, language=self.nlp.language):
                sent = Sentence(sent_id, token_id, sent, self.nlp, self.nlp.config, para_id,
                                doc_id, para_pos, doc_pos, metadata)
                para_pos, doc_pos = sent.para_pos, sent.doc_pos
                sent_list.append(sent)

                # new paragraph
                if len(sent_list) == para_size:
                    para_list.append(Paragraph(para_id, sent_list, self.nlp.config, metadata))
                    sent_list = list()
                    # next paragraph
                    para_id += 1

                # next token
                token_id = sent.token_id
                # next sentence
                sent_id += 1

            # Reset sentence list. Remove item from the previous source.
            # The last paragraph can be shorter than the others.
            # new paragraph
            if len(sent_list) > 0:
                para_list.append(Paragraph(para_id, sent_list, self.nlp.config, metadata))
                para_id += 1
                sent_list = list()

            # new document
            doc = Document(doc_id, para_list, self.nlp.config, metadata)
            para_list = list()

            # next document
            doc_id += 1

            # build records
            doc_records, para_records, sent_records, token_records = list(), list(), list(), list()
            if index_doc:
                doc_records.append(doc.to_dict())
            if index_para:
                para_records.extend([para.to_dict() for para in doc.para_list])
            for para in doc.para_list:
                if index_sent:
                    sent_records.extend([sent.to_dict() for sent in para.sent_list])
                for sent in para.sent_list:
                    token_records.extend([token.to_dict() for token in sent.token_list])

            # store data in database
            tokenization_ = dict()
            tokenization_['tokens'] = token_records
            if index_sent:
                tokenization_['sentences'] = sent_records
            if index_para:
                tokenization_['paragraphs'] = para_records
            if index_doc:
                tokenization_['documents'] = doc_records

            yield tokenization_



import abc
from typing import List


class Lexicon(metaclass=abc.ABCMeta):
    """
    Lexicon-based strategies are very efficient and simple methods. They make use of a sentiment _lexicon_
     to assign a polarity value to each text document by following a basic algrithm. A sentiment lexicon is a
     list of lexical features (e.g., words, phrase, etc.) which are labeled according to their semantic
     orientation (i.e. polarity) as either _positive_ or _negative_.
    """

    @abc.abstractmethod
    def polarity(self, token: str) -> dict:
        """
        Returns a float for sentiment strength based on the input text.
        Positive values are positive valence, negative value are negative valence, neutral value are neutral valence.'

        Args:
            token (str): The input token.

        Returns:
            dict: A positive, a negative and a neutral value
        """

    def polarities(self, bag_of_words: List[str]) -> dict:
        """
        Returns the polarities of the tokens in bag_of_words.

        Args:
            bag_of_words (List[str]): List of token.

        Returns:
            dict: A positive, a negative and a neutral value for each token
        """
        return {token: self.polarity(token) for token in bag_of_words}



language = 'lithuanian'
iso639 = 'lt'

spacy_model_sm = iso639 + '_core_news_sm'
spacy_model_md = iso639 + '_core_news_md'
spacy_model_lg = iso639 + '_core_news_lg'

sentiment_lexicon_class = None
pos_token_sentiment = None

negations = ['ne', 'ne', 'nėra', 'nei', 'nei', 'niekada', 'niekas', 'nieko', 'nulis', 'nebuvimas', 'niekingumas',
       'niekis']